#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include "mpi.h"

void print_data(double**, int, int);

int main(int argc,char *argv[]) {
  int rang;          /* Rang du processus MPI */
  int nb_process;    /* Nombre de processus MPI */

  double **data ;    /* Tableau 2D contenant les donnees de calcul */
 
  int ni, nj;        /* Nombre de points suivant i et j dans data[i][j] */
  int next, prev;

  int tag = 100;      /* Etiquette associee au message */

  int i,j;

  /* Initialisation de l'environnement MPI */
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rang);
  MPI_Comm_size(MPI_COMM_WORLD,&nb_process);


  /* Initialisation de data */
  nj = 14;
  ni = (nj-2) / nb_process + 2;
  if (nb_process < 2) {
    fprintf(stderr,"Ce programme ne fonctionne qu'avec au moins 2 processus\n");
    MPI_Abort(MPI_COMM_WORLD,1);
  }

  /*** Creation du tableau 2D ***/

  /* Allocation d'un tableau de pointeurs */
  data = (double**) malloc ( ni * sizeof(double*));

  /* Allocation de la zone mémoire */
  data[0] = (double*) malloc (ni*nj * sizeof(double));

  /* Affectation des pointeurs pour chaque premier element de ligne */
  for (i=1 ; i<ni ; i++) {
    data[i] = data[i-1] + nj;
  }

  /* Initialisation des valeurs du tableau */
  for (i=1 ; i<ni-1 ; i++) {
    for (j=1 ; j<nj-1 ; j++) {
       data[i][j] = (double) rang+10.;
    }
  }

  /* Mise a zero des points du bord */
  for (i=0 ; i<ni ; i++) {
    data[i][0] = 0.;
    data[i][nj-1] = 0.;
  }
  for (j=0 ; j<nj ; j++) {
    data[0][j] = 0.;
    data[ni-1][j] = 0.;
  }


  /*** Affichage avant communication ***/
  if (rang == 0)
      printf("Avant communication:\n");

  for (i=0 ; i<nb_process ; i++) {
    if (rang == i) {
      print_data(data,ni,nj);
    }
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
  }
  sleep(1);



  /* Echange de messages */
  next = rang+1;
  if (next == nb_process) 
     next = MPI_PROC_NULL;

  prev = rang-1;
  if (prev == -1)
     prev= MPI_PROC_NULL;

  MPI_Sendrecv(... , ..., MPI_DOUBLE, ..., tag,
               ... , ..., MPI_DOUBLE, ..., tag,
               MPI_COMM_WORLD, MPI_STATUS_IGNORE);
  MPI_Sendrecv(... , ..., MPI_DOUBLE, ..., tag,
               ... , ..., MPI_DOUBLE, ..., tag, 
               MPI_COMM_WORLD, MPI_STATUS_IGNORE);


  /*** Affichage apres communication ***/
  sleep(1);
  MPI_Barrier(MPI_COMM_WORLD);
  if (rang == 0)
    printf("Apres communication:\n");
  MPI_Barrier(MPI_COMM_WORLD);
  fflush(stdout);
  for (i=0 ; i<nb_process ; i++) {
    if (rang == i) {
      print_data(data,ni,nj);
    }
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
  }


  /* Fermeture de l'environnement MPI */
  MPI_Finalize();
  return 0;
}

void print_data(double** data, int ni, int nj) {
  int i,j;

  printf("......");
  for (j=0 ; j<nj ; j++)
    printf("%4.1f ", data[0][j]);
  printf("\n");
  for (i=1 ; i<ni-1 ; i++) {
    printf("      ");
    for (j=0 ; j<nj ; j++)
      printf("%4.1f ", data[i][j]);
    printf("\n");
  }
  printf("......");
  for (j=0 ; j<nj ; j++)
    printf("%4.1f ", data[ni-1][j]);
  printf("\n\n");
}


