#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include "mpi.h"

void print_data(double**, int, int);

int main(int argc,char *argv[]) {
  int rang;          /* Rang du processus MPI */
  int nb_process;    /* Nombre de processus MPI */

  double **data ;    /* Tableau 2D contenant les donnees de calcul */
  double *buffer_next;    /* Zone tampon pour l'echange de donnees */
  double *buffer_prev;    /* Zone tampon pour l'echange de donnees */
 
  int ni, nj;        /* Nombre de points suivant i et j dans data[i][j] */
  int next, prev;

  int tag = 100;      /* Etiquette associee au message */

  int i,j;

  /* Initialisation de l'environnement MPI */
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD,&rang);
  MPI_Comm_size(MPI_COMM_WORLD,&nb_process);


  /* Initialisation de data */
  nj = 12;
  ni = nj / nb_process;
  if (nb_process < 2) {
    fprintf(stderr,"Ce programme ne fonctionne qu'avec au moins 2 processus\n");
    MPI_Abort(MPI_COMM_WORLD,1);
  }

  /*** Creation du tableau 2D ***/

  /* Allocation d'un tableau de pointeurs */
  data = (double**) malloc ( ni * sizeof(double*));

  /* Allocation de la zone mémoire */
  data[0] = (double*) malloc (ni*nj * sizeof(double));

  /* Affectation des pointeurs pour chaque premier element de ligne */
  for (i=1 ; i<ni ; i++) {
    data[i] = data[i-1] + nj;
  }

  /* Initialisation des valeurs du tableau */
  for (i=0 ; i<ni ; i++) {
    for (j=0 ; j<nj ; j++) {
       data[i][j] = (double) rang;
    }
  }



  /*** Affichage avant communication ***/
  if (rang == 0)
      printf("Avant communication:\n");

  for (i=0 ; i<nb_process ; i++) {
    if (rang == i) {
      print_data(data,ni,nj);
    }
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
  }



  /* Allocation d'une zone tampon pour l'echange */
  buffer_next = (double*) malloc (nj * sizeof(double));
  buffer_prev = (double*) malloc (nj * sizeof(double));
  for (j=0 ; j<nj ; j++) {
    buffer_prev[j] = (double) rang;
    buffer_next[j] = (double) rang;
  }

  next = rang+1;
  if (next == nb_process) 
     next = MPI_PROC_NULL;

  prev = rang-1;
  if (prev == -1)
     prev= MPI_PROC_NULL;

  /* Copie de la bande dans le tampon et echange de messages */
  MPI_Sendrecv(data[ni-1] , nj, MPI_DOUBLE, next, tag, 
               buffer_next, nj, MPI_DOUBLE, next, tag, 
               MPI_COMM_WORLD, MPI_STATUS_IGNORE);
  MPI_Sendrecv(data[0]    , nj, MPI_DOUBLE, prev, tag,
               buffer_prev, nj, MPI_DOUBLE, prev, tag, 
               MPI_COMM_WORLD, MPI_STATUS_IGNORE);

  for (j=0 ; j<nj ; j++) {
      data[0][j]    = buffer_prev[j];
      data[ni-1][j] = buffer_next[j];
  }


  /*** Affichage apres communication ***/
  sleep(1);
  MPI_Barrier(MPI_COMM_WORLD);
  if (rang == 0)
    printf("Apres communication:\n");
  MPI_Barrier(MPI_COMM_WORLD);
  fflush(stdout);
  for (i=0 ; i<nb_process ; i++) {
    if (rang == i) {
      print_data(data,ni,nj);
    }
    fflush(stdout);
    MPI_Barrier(MPI_COMM_WORLD);
  }


  /* Fermeture de l'environnement MPI */
  MPI_Finalize();
  return 0;
}

void print_data(double** data, int ni, int nj) {
  int i,j;

  for (i=0 ; i<ni ; i++) {
    printf("      ");
    for (j=0 ; j<nj ; j++)
      printf("%4.1f ", data[i][j]);
    printf("\n");
  }
  printf("\n");
}


